#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Convert 15+1-bit RRRRRRGGGGGBBBBBA format to 32-bit ARGB format!
static void convert_palette(unsigned char *src, unsigned int *palette, int len)
{
	int i;

	for (i = 0; i < len * 2; i+=2)
	{
		unsigned char r,g,b,alpha;
		unsigned short d = src[i + 1] * 256 + src[i + 0];

		r = (d >> 11) & 0x1f;
		g = (d >> 6) & 0x1f;
		b = (d >> 1) & 0x1f;
		alpha = (d & 1) ? 0xff : 0;

		r = (r << 3) | (r >> 2);
		g = (g << 3) | (g >> 2);
		b = (b << 3) | (b >> 2);

		palette[i/2] = (alpha << 24) | (r << 16) | (g << 8) | b;
	}
}

// Convert 4 bit-per-pixel packed format (with color table) to 24-bit BGR format
static void do_04bits(unsigned char *src, unsigned char *dst, unsigned char *alp, int w, int h)
{
	int i,j;
	unsigned int palette[0x10];

	convert_palette(src, palette, 16);

	src += 16 * 2; // skip color table

	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j+=2, dst+=6, alp+=6) {
			unsigned int t = ((((h - 1) - i) * w) + j) / 2;

			dst[0] = palette[src[t]>>4] >> 0;
			dst[1] = palette[src[t]>>4] >> 8;
			dst[2] = palette[src[t]>>4] >> 16;

			alp[0] = alp[1] = alp[2] = palette[src[t]>>4] >> 24; // alpha

			dst[3] = palette[src[t]&0xf] >> 0;
			dst[4] = palette[src[t]&0xf] >> 8;
			dst[5] = palette[src[t]&0xf] >> 16;

			alp[3] = alp[4] = alp[5] = palette[src[t]>>4] >> 24; // alpha
		}
	}
}

// Convert 8 bit-per-pixel packed format (with color table) to 24-bit BGR format
static void do_08bits(unsigned char *src, unsigned char *dst, unsigned char *alp, int w, int h)
{
	int i,j;
	unsigned int palette[0x100];

	convert_palette(src, palette, 256);

	src += 256 * 2; // skip color table

	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++, dst+=3, alp+=3) {
			unsigned int t = ((((h - 1) - i) * w) + j) * 1;

			unsigned int pxl = palette[src[t]];

			dst[0] = pxl & 0xff;
			dst[1] = pxl >> 8;
			dst[2] = pxl >> 16;

			alp[0] = alp[1] = alp[2] = pxl >> 24; // alpha
		}
	}
}

// Convert 15+1-bit RRRRRRGGGGGBBBBBA format to 24-bit BGR format!
static void do_16bits(unsigned char *src, unsigned char *dst, unsigned char *alp, int w, int h)
{
	int i,j;

	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++, dst+=3, alp+=3) {
			unsigned int t = ((((h - 1) - i) * w) + j) * 2;

			unsigned char r,g,b,alpha;
			unsigned short d = src[t + 1] * 256 + src[t + 0];

			r = (d >> 11) & 0x1f;
			g = (d >> 6) & 0x1f;
			b = (d >> 1) & 0x1f;
			alpha = (d & 1) ? 0xff : 0; // unused for now...

			r = (r << 3) | (r >> 2);
			g = (g << 3) | (g >> 2);
			b = (b << 3) | (b >> 2);

			dst[0] = b;
			dst[1] = g;
			dst[2] = r;

			alp[0] = alp[1] = alp[2] = alpha; // alpha
		}
	}
}

// Convert 24-bit RGB to 24-bit BGR format!
static void do_24bits(unsigned char *src, unsigned char *dst, unsigned char *alp, int w, int h)
{
	int i,j;

	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++, dst+=3, alp+=3) {
			unsigned int t = ((((h - 1) - i) * w) + j) * 3;

			dst[0] = src[t + 2];
			dst[1] = src[t + 1];
			dst[2] = src[t + 0];


			alp[0] = alp[1] = alp[2] = 0xff; // no alpha!!
		}
	}
}

// Convert 32-bit RGBA format to 24-bit BGR format!
static void do_32bits(unsigned char *src, unsigned char *dst, unsigned char *alp, int w, int h)
{
	int i,j;

	for (i = 0; i < h; i++) {
		for (j = 0; j < w; j++, dst+=3, alp+=3) {
			unsigned int t = ((((h - 1) - i) * w) + j) * 4;

			alp[0] = alp[1] = alp[2] = src[t + 3]; // alpha

			dst[0] = src[t + 2];
			dst[1] = src[t + 1];
			dst[2] = src[t + 0];
		}
	}
}

int main(int argc, char **argv)
{
	char name[256];

	if (argc < 2) {
		printf ("File name:\n");
		gets(name);
	} else {
		strcpy (name, argv[1]);
	}

	FILE *fa = fopen(name,"rb");

	if (fa == NULL) {
		printf ("%s not found!\n", name);
		return 1;
	}

	fseek (fa, 0, SEEK_END);

	int len = ftell (fa);

	rewind (fa);

	unsigned char *src = (unsigned char*)malloc(len);

	fread (src, 2, 1, fa);

	if (src[0] != 0x4f || src[1] != 0x49) { // 'OI'
		printf ("Wrong header!\n");
		free (src);
		fclose (fa);
		return 1;
	}


	fread (src+2, len-2, 1, fa);
	fclose (fa);

//	unsigned short header = src[0] + src[1] * 256; // "OI"
	unsigned short mode   = src[2] + src[3] * 256;
	unsigned short width  = src[4] + src[5] * 256;
	unsigned short height = src[6] + src[7] * 256;

	unsigned char *dst = (unsigned char*)malloc(width * height * 3);
	unsigned char *alp = (unsigned char*)malloc(width * height * 3);

	switch (mode)
	{
		case 0x0404: //  4 bits per pixel (uses  16 entry color table), color table is rrrrrgggggbbbbba format
			do_04bits(src + 8, dst, alp, width, height);
		break;

		case 0x0804: //  8 bits per pixel (uses 256 entry color table), color table is rrrrrgggggbbbbba format
			do_08bits(src + 8, dst, alp, width, height);
		break;

		case 0x2004: // 16 bits per pixel,  rrrrrgggggbbbbba format
			do_16bits(src + 8, dst, alp, width, height);
		break;

		case 0x1800: // 24 bits per pixel, rrrrrrrrggggggggbbbbbbbb format
			do_24bits(src + 8, dst, alp, width, height);
		break;

		case 0x2001: // 32 bits per pixel, rrrrrrrrggggggggbbbbbbbbaaaaaaaa format
			do_32bits(src + 8, dst, alp, width, height);
		break;

		default:
			printf ("unknown bitmap mode: %4.4x\n", mode);
		break;
	}

	{
		char outputname[256];
		unsigned char data[0x36];

		unsigned int bmp_data_offset = 0x36;
		unsigned int image_data_size = width * height * 3;
		unsigned int dst_len = bmp_data_offset + image_data_size;
		unsigned int struct_size = bmp_data_offset - 0x0e;
		unsigned short planes = 1;
		unsigned int compression = 0;
		unsigned short image_bits = 3 * 8;
		unsigned int palette_colors_used = 0; // # of colors used
		unsigned int palette_colors_important = 0; // # of colors that are important

		data[0x00] = 'B';
		data[0x01] = 'M';
		data[0x02] = dst_len >> 0;
		data[0x03] = dst_len >> 8;
		data[0x04] = dst_len >> 16;
		data[0x05] = dst_len >> 24;
		data[0x06] = 0;
		data[0x07] = 0;
		data[0x08] = 0;
		data[0x09] = 0;
		data[0x0a] = bmp_data_offset >> 0;
		data[0x0b] = bmp_data_offset >> 8;
		data[0x0c] = bmp_data_offset >> 16;
		data[0x0d] = bmp_data_offset >> 24;
		data[0x0e] = struct_size >> 0;
		data[0x0f] = struct_size >> 8;
		data[0x10] = struct_size >> 16;
		data[0x11] = struct_size >> 24;
		data[0x12] = width >> 0;
		data[0x13] = width >> 8;
		data[0x14] = width >> 16;
		data[0x15] = width >> 24;
		data[0x16] = height >> 0;
		data[0x17] = height >> 8;
		data[0x18] = height >> 16;
		data[0x19] = height >> 24;
		data[0x1a] = planes >> 0;
		data[0x1b] = planes >> 8;
		data[0x1c] = image_bits >> 0;
		data[0x1d] = image_bits >> 8;
		data[0x1e] = compression >> 0;
		data[0x1f] = compression >> 8;
		data[0x20] = compression >> 16;
		data[0x21] = compression >> 24;
		data[0x22] = image_data_size >> 0;
		data[0x23] = image_data_size >> 8;
		data[0x24] = image_data_size >> 16;
		data[0x25] = image_data_size >> 24;
		data[0x26] = 0;
		data[0x27] = 0;
		data[0x28] = 0;
		data[0x29] = 0;
		data[0x2a] = 0;
		data[0x2b] = 0;
		data[0x2c] = 0;
		data[0x2d] = 0;
		data[0x2e] = palette_colors_used >> 0;
		data[0x2f] = palette_colors_used >> 8;
		data[0x30] = palette_colors_used >> 16;
		data[0x31] = palette_colors_used >> 24;
		data[0x32] = palette_colors_important >> 0;
		data[0x33] = palette_colors_important >> 8;
		data[0x34] = palette_colors_important >> 16;
		data[0x35] = palette_colors_important >> 24;

		sprintf (outputname, "%s.bmp", name);

		fa = fopen(outputname, "wb");
		fwrite (data, 0x36, 1, fa);
		fwrite (dst, width * height * 3, 1, fa);
		fclose (fa);

		sprintf (outputname, "%s_alpha.bmp", name);

		fa = fopen(outputname, "wb");
		fwrite (data, 0x36, 1, fa);
		fwrite (alp, width * height * 3, 1, fa);
		fclose (fa);
	}

	free (alp);
	free (dst);
	free (src);

	return 0;
}
